/*
 * Beyond_Arduino 6 - 1 Temp Logger.c
 *
 *	Nuts & Volts - Beyond Arduino #6
 *  --------------------------------
 *	A functioning temperature logger:
 *   - reads temperature from LM35 sensor on PC0
 *   - reads and sets time on MCP79400 Real Time Clock (over I2C)
 *   - records time and temp on 24LC128 EEPROM (over I2C)
 *	 - interrupt-driven Serial UART interface
 *
 *  For simplicity, only works for temperatures greater than 2 degrees centigrade 
 *
 *  Connections: 
 *    - LM35 temp sensor to PC0
 *    - Microchip 24LC128 EEPROM chip: TWI (I2C)
 *    - Microchip MCP79400 RTC module: TWI (I2C)
 *
 *  EEPROM Storage structure
 *  -------------------------
 *  The data needs to be stored on the EEPROM in a format we can used, as a "record").
 *	Memory Location 1-2:
 *		- store the of the last record logged - so we do not use these.  
 *		- Functions EEPROM_getLastAddress and EEPROM_setLastAddress use these
 *	Record Structure:
 *		- 1: Year (excluding century)
 *		- 2: Month
 *		- 3: Day
 *		- 4: Hour (24-hour format)
 *		- 5: Minute
 *		- 6: Second
 *		- 7: Temperature (centigrade)
 *
 *  Author: Andrew Retallack
 *			www.crash-bang.com
 */ 




#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "RTC_MCP79400.h"
#include "EEPROM.h"
#include "uart.h"
#include "ADC.h"


#define RTC_ADDRESS 0b11011110		//I2C Address for the MCP79400 RTC.  Only first 7 bits form address.
#define EEPROM_ADDRESS 0b10101110	//I2C Address for the 24LC128 EEPROM.  A0 to A2 are high.  Only first 7 bits form address.
#define EEPROM_MAX 15500			//Maximum EEPROM address (128kbit = 16000 bytes less 500 for margin)


/*
* Variable Declarations
*/

uint16_t logInterval = 10;			//Interval at which to log temperatures (in seconds)
uint16_t intervalCounter = 0;		//Counter: counts up to the logInterval (in seconds)
volatile uint8_t uartReadByte;		//Byte read in over the UART - used by Interrupt-Handler
volatile uint8_t flagUART = 0;		//Interrupt Flag: 1 indicates data was read in from the UART - set by Interrupt-Handler
uint8_t  bTimeSet = 0;				//Is the time set?  Used to control whether logging occurs


//Time Variables- used to store the time as needed.  Not always up to date
uint8_t timeSec = 0;
uint8_t timeMin = 0;
uint8_t timeHr = 0;
uint8_t timeDay = 0;
uint8_t timeWeekDay = 0;
uint8_t timeMonth = 0;
uint16_t timeYear = 0;
uint8_t timeAmPm = 0;	//PM = 1 / AM = 0

/*
*	Function prototypes
*/
void logTemp(void);
void processUART(void);
void printLog(void);
void getUserTime(void);
uint8_t validDateElement(uint16_t testElement, uint16_t maxValue, const char elementName[]);
void printTime(void);
void writeHelpString(void);
uint8_t readTemp(void);


int main(void)
{
	uint8_t tempVar = 0; 

/************************************************************************
*  INITIALISATION
************************************************************************/

	//Initialise the UART: 9600 / 8 / N / 1
	UART_Init(9600);
	UART_writeString("Temperature Logger\r\n");

	//Initialise the ADC on PC0 (temperature sensor)
	ADC_init(0);
	
	//Initialise the I2C Interface at 200kHz
	I2C_init(200);
	
	//Initialise the RTC.  Returns a "1" if the oscillator started successfully
	tempVar = RTC_Init(RTC_ADDRESS, 1, 1);
	if (tempVar==0)
	{
		UART_writeString("ERROR: Oscillator did NOT start\r\n");
	}

/************************************************************************
*  TIME INITIALISATION
************************************************************************/

	//Check whether time has been set or not, and flag it.  (We expect time NOT to be set, but this function accommodates backup battery)
	//Read the Time
	tempVar = RTC_GetTime(RTC_ADDRESS, &timeYear, &timeMonth, &timeDay, &timeWeekDay, &timeHr, &timeAmPm, &timeMin, &timeSec);

	bTimeSet = 1;
	//If Year is one and month is one and day is one and hour is zero, assume time not set - these are the Power-On-Reset values
	if ((timeYear == 1) && (timeMonth == 1) && (timeDay == 1) && (timeMin == 0))
	{
		bTimeSet = 0;	//Time not set
	}

	
/************************************************************************
*  OTHER VARIOUS INITIALISATION
************************************************************************/

//Double the log interval: we will delay for 500ms (i.e. 0.5 sec) each loop, therefore need to count to double the configured interval
logInterval = logInterval * 2;

//Initialise Interrupt: UART Receive
UCSR0B |= (1<<RXCIE0);

//Enable Global Interrupts
sei();


while(1)
{
	
	//Is it time to Log yet?
	if (intervalCounter >= logInterval)
	{
		if  (bTimeSet == 1)	//only actually log if the time has been set
		{
			logTemp();
		}
		
		intervalCounter = 0;	//Reset interval counter
	}
	

	//Is there an incoming character on UART?
	if (flagUART)
	{
		processUART();	//Handle the UART menu functionality
	}
	
	
	//Increase interval counter
	intervalCounter++;
	_delay_ms(500);
	
}

}




/***********************************************************************
*
* Routine to Log a temperature entry.
*		- Reads the Time
*		- Reads the Temperature
*		- Stores to EEPROM
*
* INPUT:	none
* RETURNS:	none
*
************************************************************************/
void logTemp(void)
{
	uint8_t readTempVal = 0;		//Stores Temperature
	uint16_t memoryLocation = 0;	//Current EEPROM memory location being written to
	
	//Read the Time from the RTC
	readTempVal = RTC_GetTime(RTC_ADDRESS, &timeYear, &timeMonth, &timeDay, &timeWeekDay, &timeHr, &timeAmPm, &timeMin, &timeSec);
	
	//Read the Temperature
	readTempVal = readTemp();
	
	//Where will we store this?  New address = Last Address used + 1
	memoryLocation = EEPROM_getLastAddress(EEPROM_ADDRESS) + 1;
	
	//Write the elements to EEPROM in order, incrementing memory location after each
	EEPROM_write(EEPROM_ADDRESS, memoryLocation, timeYear);
	memoryLocation++;
	EEPROM_write(EEPROM_ADDRESS, memoryLocation, timeMonth);
	memoryLocation++;
	EEPROM_write(EEPROM_ADDRESS, memoryLocation, timeDay);
	memoryLocation++;
	EEPROM_write(EEPROM_ADDRESS, memoryLocation, timeHr);
	memoryLocation++;
	EEPROM_write(EEPROM_ADDRESS, memoryLocation, timeMin);
	memoryLocation++;
	EEPROM_write(EEPROM_ADDRESS, memoryLocation, timeSec);
	memoryLocation++;
	EEPROM_write(EEPROM_ADDRESS, memoryLocation, readTempVal); //Finally, write the temperature
	
	//Let us know what we logged over the UART
	UART_writeString("Logging temp: ");
	UART_printDecimal(readTempVal,0);
	UART_writeString(" at EEPROM addr ");
	UART_printDecimal(memoryLocation - 7,0);	//subtract 7 to get to first memory location
	UART_writeString(" time: ");
	printTime();

	//Check we aren't past the last "safe" address on our 128k EEPROM module
	if (memoryLocation > EEPROM_MAX)
	{
		UART_writeString("Past last EEPROM Address: reset\r\n");
		memoryLocation = 2;
	}


	//Store the last address written to, for next time
	EEPROM_setLastAddress(EEPROM_ADDRESS, memoryLocation);
	
	//All done!
}




/***********************************************************************
*
* Routine to handle the processing once a byte is received over UART:
*	- ?: Show help message
*   - T: Set the time
*   - X: Clear the log
*	- P: Print the log
*
* INPUT:	none
* RETURNS:	none
************************************************************************/
void processUART(void)
{

	//Disable Global Interrupts so we can process this in peace
	cli();
	
	//Action based on which character received over the UART
	switch (uartReadByte)
	{
		
		case '?':	//User needs help
		writeHelpString();
		break;
		
		
		case 'T':	//User wants to set the time
		case 't':
		getUserTime();
		break;
		
		
		case 'X':	//User wants to clear the Log
		case 'x':
		EEPROM_setLastAddress(EEPROM_ADDRESS, 2);	//We don't actually erase the logged data, just reset the logging to first address
		UART_writeString("Log Cleared\r\n");
		break;
		
		
		case 'P':	//User wants to print the log
		case 'p':
		printLog();
		break;
		
		
		default:	//We're not sure WHAT the user wants to do - invalid key
		UART_writeString("Unrecognised command. Press ? for help.\r\n");
		break;
	}

	flagUART = 0;	//Reset UART interrupt flag

	//Re-enable Global Interrupts
	sei();

}


/***********************************************************************
*
* This routine prints the Log out over the UART, in CSV format for easy
* import into spreadsheet
*
* INPUT:	none
* RETURNS:	none
************************************************************************/
void printLog(void)
{
	uint16_t currAddress = 3;	//First EEPROM Address
	uint16_t lastAddress = 0;	//Last EEPROM Address written to
	uint8_t readValue = 0;		//Value read from EEPROM
	
	//Get the last address written to
	lastAddress = EEPROM_getLastAddress(EEPROM_ADDRESS);
	
	//WRite UART Header
	UART_writeString("\r\nLogged Data");
	UART_writeString("\r\n-----------\r\n");
	
	//Column Headings
	UART_writeString("Date,Time,Temperature\r\n");
	

	//Loop through the EEPROM until we reach the last address logged to
	while (currAddress < lastAddress)
	{

		//Read the logged data in sequence: year, month, day, hour, min, sec, temp

		//Year
		readValue = EEPROM_read(EEPROM_ADDRESS, currAddress);	//Read From EEPROM
		UART_printDecimal(20,0);		//Print the century
		UART_printDecimal(readValue,2);	//Print the year
		UART_writeChar('/');	//Date separator
		currAddress++;	//Move onto next EEPROM memory location
		
		//Month
		readValue = EEPROM_read(EEPROM_ADDRESS, currAddress);	//Read From EEPROM
		UART_printDecimal(readValue,2);	//Print the month
		UART_writeChar('/');	//Date separator
		currAddress++;	//Move onto next EEPROM memory location
		
		//Day
		readValue = EEPROM_read(EEPROM_ADDRESS, currAddress);	//Read From EEPROM
		UART_printDecimal(readValue,2);	//Print the day
		currAddress++;	//Move onto next EEPROM memory location
		UART_writeChar(',');	//Comma-separated
		
		//Hour
		readValue = EEPROM_read(EEPROM_ADDRESS, currAddress);	//Read From EEPROM
		UART_printDecimal(readValue,2);	//Print the Hour
		UART_writeChar(':');	//Time separator
		currAddress++;	//Move onto next EEPROM memory location
		
		//Minute
		readValue = EEPROM_read(EEPROM_ADDRESS, currAddress);	//Read From EEPROM
		UART_printDecimal(readValue,2);	//Print the minute
		UART_writeChar(':');	//Time separator
		currAddress++;	//Move onto next EEPROM memory location

		//Second
		readValue = EEPROM_read(EEPROM_ADDRESS, currAddress);	//Read From EEPROM
		UART_printDecimal(readValue,2);	//Print the seconds
		currAddress++;	//Move onto next EEPROM memory location
		UART_writeChar(',');	//Comma-separated

		//Temperature
		readValue = EEPROM_read(EEPROM_ADDRESS, currAddress);	//Read From EEPROM
		UART_printDecimal(readValue,0);	//Print the temperature
		currAddress++;	//Move onto next EEPROM memory location
		
		UART_writeString("\r\n");	//Start a New line
		
	}
	
	//All Done!
}



/***********************************************************************
*
* This routine reads the time entered over the UART, validates it,
* and finally sets the RTC
*
* INPUT:	none
* RETURNS:	none
************************************************************************/
void getUserTime(void)
{
	
	uint8_t	readByte = 0;	//Byte read in over UART
	
	//The user has pressed the "T" key to set the time, now we capture the
	//time in the following format:
	//YYMMDD	Year (no Century), month, day
	//HHMMSS	Hours (24-hour format), minutes, seconds
	
	//Year
	readByte = UART_readChar_Wait(5) - 48;	//Read a byte in, with a 5 second timeout, converted from ASCII to "0-9"
	timeYear = readByte * 10;	//first Tens
	readByte = UART_readChar_Wait(5) - 48;	//Read a byte in, with a 5 second timeout, converted from ASCII to "0-9"
	timeYear += readByte;	//then Units
	if (validDateElement(timeYear, 99, "YEAR") == 0)	//Validate: Year is <= 99.  Failure means we exit.
	{
		return;
	}
	
	//Month
	readByte = UART_readChar_Wait(5) - 48;	//Read a byte in, with a 5 second timeout, converted from ASCII to "0-9"
	timeMonth = readByte * 10;	//first Tens
	readByte = UART_readChar_Wait(5) - 48;	//Read a byte in, with a 5 second timeout, converted from ASCII to "0-9"
	timeMonth += readByte;	//then Units
	if (validDateElement(timeMonth, 12, "MONTH") == 0)	//Validate: Month is <= 12.  Failure means we exit.
	{
		return;
	}

	//Day
	readByte = UART_readChar_Wait(5) - 48;	//Read a byte in, with a 5 second timeout, converted from ASCII to "0-9"
	timeDay = readByte * 10;	//first Tens
	readByte = UART_readChar_Wait(5) - 48;	//Read a byte in, with a 5 second timeout, converted from ASCII to "0-9"
	timeDay += readByte;	//then Units
	if (validDateElement(timeDay, 31, "DAY") == 0)	//Validate: Day is <= 31.  Failure means we exit.
	{
		return;
	}

	//Hours
	readByte = UART_readChar_Wait(5) - 48;	//Read a byte in, with a 5 second timeout, converted from ASCII to "0-9"
	timeHr = readByte * 10;	//first Tens
	readByte = UART_readChar_Wait(5) - 48;	//Read a byte in, with a 5 second timeout, converted from ASCII to "0-9"
	timeHr += readByte;	//then Units
	if (validDateElement(timeHr, 23, "HOUR") == 0)	//Validate: Hours is <= 23.  Failure means we exit.
	{
		return;
	}

	//Minutes
	readByte = UART_readChar_Wait(5) - 48;	//Read a byte in, with a 5 second timeout, converted from ASCII to "0-9"
	timeMin = readByte * 10;	//first Tens
	readByte = UART_readChar_Wait(5) - 48;	//Read a byte in, with a 5 second timeout, converted from ASCII to "0-9"
	timeMin += readByte;	//then Units
	if (validDateElement(timeMin, 59, "MINUTE") == 0)	//Validate: Minute is <= 59.  Failure means we exit.
	{
		return;
	}

	//Seconds
	readByte = UART_readChar_Wait(5) - 48;	//Read a byte in, with a 5 second timeout, converted from ASCII to "0-9"
	timeSec = readByte * 10;	//first Tens
	readByte = UART_readChar_Wait(5) - 48;	//Read a byte in, with a 5 second timeout, converted from ASCII to "0-9"
	timeSec += readByte;	//then Units
	if (validDateElement(timeSec, 59, "SECOND") == 0)	//Validate: Second is <= 59.  Failure means we exit.
	{
		return;
	}


	//Time has been successfully captured.  Now set the RTC
	
	readByte = RTC_SetTime(RTC_ADDRESS, timeYear, timeMonth, timeDay, 1, timeHr, 0, timeMin, timeSec);	//We don't use Day of Week or AM/PM indicator
	
	//The RTC_SetTime function returns a "1" if successful
	if (readByte)
	{
		UART_writeString("\r\nSuccess.  Time set: ");
		printTime();
		bTimeSet = 1;	//We can now start logging, if we weren't already
	}

	
}


/***********************************************************************
*
* Helper routine: Validates a number falls below a max, otherwise
* writes error to UART
*
* INPUT:	testElement:	Value to test for validity
*			maxValue:		Maximum value the test value can be
*			elementName:	Descriptor of element to be written to UART error
* RETURNS:	Is value in range?  1 = valid / 0 = not valid
************************************************************************/

uint8_t validDateElement(uint16_t testElement, uint16_t maxValue, const char elementName[])
{
	if (testElement > maxValue)	//If value is not valid, print error message
	{
		UART_writeString("The ");
		UART_writeString(elementName);
		UART_writeString(" is invalid. Max value is: ");
		UART_printDecimal(maxValue, 0);
		UART_writeString(" / you entered: ");
		UART_printDecimal(testElement, 0);
		UART_writeString("\r\n");
		
		return 0;	//Failure
	}
	
	return 1;	//Passed validation
}



/***********************************************************************
*
* Routine to print time over the UART in a readable format
*
* INPUT:	none
* RETURNS:	none
*
************************************************************************/

void printTime(void)
{
	
	//Output date over the UART: dd/MM/yyyy - that's how we do it in South Africa :)
	//All elements padded with zeros to two characters (eg. if hour is "1", output "01")
	UART_printDecimal(timeDay,2);
	UART_writeString("/");
	UART_printDecimal(timeMonth,2);
	UART_writeString("/20");
	UART_printDecimal(timeYear,2);
	
	//Output time over the UART: HH:mm:ss, 24-hour format :)
	UART_writeString("   ");
	UART_printDecimal(timeHr,2);
	UART_writeString(":");
	UART_printDecimal(timeMin,2);
	UART_writeString(":");
	UART_printDecimal(timeSec,2);
	UART_writeString("\r\n");

}


/***********************************************************************
*
* This routine writes the help message out
*
* INPUT:	none
* RETURNS:	none
************************************************************************/
void writeHelpString(void)
{
	uint8_t result = 0;
	
	//Read the Time
	result = RTC_GetTime(RTC_ADDRESS, &timeYear, &timeMonth, &timeDay, &timeWeekDay, &timeHr, &timeAmPm, &timeMin, &timeSec);
	
	UART_writeString("Temperature Logger\r\n");
	UART_writeString("------------------\r\n");
	
	if (bTimeSet == 0)
	{
		UART_writeString("Time not set.");
	}
	else
	{
		UART_writeString("Current Time: ");
		printTime();	//printTime uses the time we read in earlier, accessing the global variables
	}
	
	UART_writeString("\r\nCommands:\r\n");
	UART_writeString("  ?: Display help\r\n");
	UART_writeString("  Tyymmddhhmmss: Set date and time:\r\n");
	UART_writeString("      year, month, day, hour, minutes, seconds \r\n\r\n");
	UART_writeString("  X: Clear Log\r\n");
	UART_writeString("  P: Print Log\r\n\r\n");

}



/***********************************************************************
*
* This routine reads the temperature sensor and returns current temperature
*
* INPUT:	none
* RETURNS:	the temperature in centigrade
************************************************************************/
uint8_t readTemp(void)
{
	
	uint16_t readingADC;
	
	//Read the ADC value (0 - 1023)
	readingADC = ADC_read();
	
	//  Use below for debugging if needed
	//UART_writeString("ADC Value: ");
	//UART_printDecimal(readingADC,0);
	//UART_writeString("\r\n");
	
	
	//Convert to a millivoltage (we're operating at 5V)
	readingADC = (readingADC * 5000UL) / 1024UL;	//Need to cast the constant numbers as unsigned long to ensure correct calculation
	
	//Return the temperature: 10mV per degree, with a base offset of 0mV = 2 deg
	return (readingADC / 10) + 2;
	
}


/***********************************************************************
*
* This is the ISR (interrupt service routine) for UART_RX
*   It is triggered when a byte is received over the UART
*
************************************************************************/
ISR(USART_RX_vect)
{
	uartReadByte = UDR0;	//Read the received byte from the UART
	flagUART = 1;	//Set the flag so the main routine knows to handle button press
}

