/*
 * @file	RTC_MCP79400.c
 *
 *  ------------------------------------
 *  @author	Andrew Retallack, Crash-Bang Prototyping
 *			www.crash-bang.com
 *  @date	30/05/2015
 *
 */ 

#include "RTC_MCP79400.h"


/*
*	Variables storing settings
*/
volatile uint8_t settingBackupBat = 1;	//Is Backup Battery Enabled?
volatile uint8_t setting24Hour = 1;		//Do we use 24-hour format?


/**
* @brief	
*
* @details	
*			
*
* @param[in]	
*
* @return	
************************************************************************/


/**
* @brief		Initialise the RTC
*
* @details		Initialise the RTC: Oscillator running. 
*				Record the hour format (24-hr vs 12-hr) - does not actually set this in the RTC
*				Record whether backup battery connected - does not actually set this in the RTC
*
*
* @param[in]	deviceAddress	The I2C address of the RTC device
*				is24Hour		Are we using 24-hour format (1=24 / 0 = 12)
*				isBackupBat		Is a backup battery connected (1=Yes)?
*
* @return	Oscillator state: 1=running / 0=not started
************************************************************************/
uint8_t RTC_Init(uint8_t deviceAddress, uint8_t is24Hour, uint8_t isBackupBat)
{
	
	uint8_t tempVar = 0;
	
	//Store settings in module-level variable
	settingBackupBat = isBackupBat;
	setting24Hour = is24Hour;
	
	
	//Read Oscillator and Backup Bat settings from the MCP79400 - both are in register "RTCWKDAY"
	tempVar = RTC_read(deviceAddress, MCP794_RTCWKDAY);

	//Backup Battery: If setting on module is not same as setting passed, update it.
	// (we need to read full register in, amend it, then rewrite back
	if ((tempVar & (1<<MCP794_VBATEN)) != isBackupBat)
	{
		if (isBackupBat == 0)
		{
			tempVar &= ~(1<<MCP794_VBATEN);	//Disable the backup battery bit
		}
		else
		{
			tempVar |= (1<<MCP794_VBATEN);	//Enable the backup battery bit
		}
		
		RTC_write(deviceAddress, MCP794_RTCWKDAY, tempVar);	//Write setting back
		UART_writeString("Backup Battery: ");
		UART_printDecimal(tempVar,0);
		UART_writeString("\r\n");
	}
	
	
	//Oscillator: if not running, start it.  
	// (If the RTC was already running on backup battery, the Osc will already be running)
	if ( (tempVar & (1<<MCP794_OSCRUN)) == 0)
	{
		RTC_write(deviceAddress, MCP794_CONTROL, 0);	//MCP79400 can run off external oscillater - we do not want this so ensure disabled
		RTC_write(deviceAddress, MCP794_RTCSEC, 30 | (1<<MCP794_ST));	//Reset seconds to 30 (arbitrary) and start oscillator
	}
	
	_delay_ms(10);	//Give oscillator plenty of time to start up
	
	//Check the Oscillator Started
	tempVar = RTC_read(deviceAddress, MCP794_RTCWKDAY);
	if ( (tempVar & (1<<MCP794_OSCRUN)) == 0)	//Oscillator is not running
	{
		return 0;	//Error: Oscillator didn't start
	}
	
	return 1;	//All OK, oscillator started
}



/**
* @brief	Set the time on the the RTC
*
* @details	This function sets the time
*
* @param[in]	deviceAddress	The I2C address of the RTC device
* @param[in]	setXxxxx		The Value of each Date/Time Element to set
*
* @return	1 = Success
************************************************************************/

uint8_t RTC_SetTime(uint8_t deviceAddress, uint16_t setYear, uint8_t setMonth, uint8_t setDay, uint8_t setWeekDay, uint8_t setHour, uint8_t isHourPM, uint8_t setMinutes, uint8_t setSeconds)
{
	uint8_t result;
	uint8_t sendData = 0;
	

	//Disable Oscillator to prevent time clocking over into new hour/day/month etc. during setting
	RTC_write(deviceAddress, MCP794_RTCSEC, 0);  //This also results in seconds going to zero - does not matter as we set them later
	
	//Send START Condition
	result = I2C_sendStart();


	//---Year
	result = I2C_send(deviceAddress|TW_WRITE);  //Send the device Address with WRITE
	result = I2C_send(MCP794_RTCYEAR);			//Send Year Register address
	//Only interested in last 2 digits of year:
	sendData = setYear;
	if (setYear > 99)
	{
		sendData = setYear % 100;
	}

	result = I2C_send(decToBcd(sendData));	//Send Year Value in BCD


	//---Month
	result = I2C_sendStart();
	result = I2C_send(deviceAddress|TW_WRITE);  //Send the device Address with WRITE
	result = I2C_send(MCP794_RTCMTH);			//Send Month Register Address
	result = I2C_send(decToBcd(setMonth));		//Send Month Value in BCD


	//---Day
	result = I2C_sendStart();
	result = I2C_send(deviceAddress|TW_WRITE);  //Send the device Address with WRITE
	result = I2C_send(MCP794_RTCDATE);			//Send Day Register Address
	result = I2C_send(decToBcd(setDay));		//Send Day Value in BCD

	//---WeekDay
	result = I2C_sendStart();
	result = I2C_send(deviceAddress|TW_WRITE);  //Send the device Address with WRITE
	result = I2C_send(MCP794_RTCWKDAY);			//Send Day Register Address
	result = I2C_send(decToBcd(setWeekDay) | (settingBackupBat<<MCP794_VBATEN));	//Send WeekDay Value in BCD, OR'ed with the Backup Battery Setting
																					// (We need to set entire register value, not just the parts we wish)

	//---Hour
	result = I2C_sendStart();
	result = I2C_send(deviceAddress|TW_WRITE);  //Send the device Address with WRITE
	result = I2C_send(MCP794_RTCHOUR);			//Send the Register Address
	
	sendData = setHour;
	if (setting24Hour)	//If 24 Hour format:
	{
		result = I2C_send(decToBcd(sendData));		//Send Hour Value in BCD - also sets time to 24-hr mode (bit 6 of register = "0")
	}
	else  //12-hour format
	{
		if (sendData > 12)	//Check the hour parameter is actually 12-hr format
		{
			sendData = sendData - 12;  //convert to 12-hr
		}
		result = I2C_send(decToBcd(sendData) | (1 << MCP794_12_24) | (isHourPM << MCP794_AM_PM));	//Send Hour Value in BCD - also set AM/PM and 12-hour format
	}
	

	
	//---Minutes
	result = I2C_sendStart();
	result = I2C_send(deviceAddress|TW_WRITE);  //Send the device Address with WRITE
	result = I2C_send(MCP794_RTCMIN);			//Send Minute Register Address
	result = I2C_send(decToBcd(setMinutes));	//Send Minute Value in BCD
	

	//---Seconds
	result = I2C_sendStart();
	result = I2C_send(deviceAddress|TW_WRITE);  //Send the device Address with WRITE
	result = I2C_send(MCP794_RTCSEC);			//Send Second Register Address
	result = I2C_send(decToBcd(setSeconds) | (1<<MCP794_ST));	//Send Second Value in BCD - Also start oscillator


	//Send STOP Condition
	I2C_sendStop();

	_delay_ms(10);	//Give oscillator plenty of time to start


	//Check the Oscillator Started
	result = RTC_read(deviceAddress, MCP794_RTCWKDAY);
	if ( (result & (1<<MCP794_OSCRUN)) == 0)	//Oscillator is not running
	{
		return 0;	//Error: Oscillator didn't start
	}
	
	return 1;	//All OK, oscillator started
	
}



/**
* @brief	Read the time on the the RTC
*
* @details	This function reads the time
*
* @param[in]	deviceAddress	The I2C address of the RTC device
* @param[in]	getXxxxx		Pointer to each Date/Time Element to retrieve
*
* @return	1 = Success
************************************************************************/

uint8_t RTC_GetTime(uint8_t deviceAddress, uint16_t *getYear, uint8_t *getMonth, uint8_t *getDay, uint8_t *getWeekDay, uint8_t *getHour, uint8_t *isHourPM, uint8_t *getMinutes, uint8_t *getSeconds)
{
	uint8_t result;
	uint8_t readData = 0;
	
	//We need to retrieve the entire date/time in one read, so as to prevent time clocking over
	
	//Send START Condition.
	result = I2C_sendStart();
	
	//---Year
	result = I2C_send(deviceAddress|TW_WRITE);  //Send the device Address with WRITE
	result = I2C_send(MCP794_RTCYEAR);			//Send Year Register Address - to read from
	result = I2C_sendStart();					//Repeated Start - now we're going to do a read operation
	result = I2C_send(deviceAddress|TW_READ);	//Send the device Address with READ
	readData = I2C_read(0);						//Read the value from the register
	*getYear = bcdToDec(readData);				//Convert from BCD and store value in pointer
	


	//---Month
	result = I2C_sendStart();					//Repeated Start - now we're going to do a Write operation
	result = I2C_send(deviceAddress|TW_WRITE);  //Send the device Address with WRITE
	result = I2C_send(MCP794_RTCMTH);			//Send Month Register Address - to read from
	result = I2C_sendStart();					//Repeated Start - now we're going to do a read operation
	result = I2C_send(deviceAddress|TW_READ);	//Send the device Address with READ
	readData = I2C_read(0) & MCP794_MASK_Month;	//Read the value from the register
	*getMonth = bcdToDec(readData);				//Convert from BCD and store value in pointer


	//---Day
	#if DEBUGLEVEL > 2
	UART_writeString("---Day---\r\n");
	#endif
	result = I2C_sendStart();
	result = I2C_send(deviceAddress|TW_WRITE);  //Send the device Address with WRITE
	result = I2C_send(MCP794_RTCDATE);			//Send Day Register Address - to read from
	result = I2C_sendStart();					//Repeated Start - now we're going to do a read operation
	result = I2C_send(deviceAddress|TW_READ);	//Send the device Address with READ
	readData = I2C_read(0);						//Read the value from the register
	*getDay = bcdToDec(readData);				//Convert from BCD and store value in pointer


	//---Hour
	#if DEBUGLEVEL > 2
	UART_writeString("---Hour---\r\n");
	#endif
	result = I2C_sendStart();
	result = I2C_send(deviceAddress|TW_WRITE);  //Send the device Address with WRITE
	result = I2C_send(MCP794_RTCHOUR);			//Send Hour Register Address - to read from
	result = I2C_sendStart();					//Repeated Start - now we're going to do a read operation
	result = I2C_send(deviceAddress|TW_READ);	//Send the device Address with READ
	readData = I2C_read(0);						//Read the value from the register
	
	if (readData & (1<<MCP794_12_24))	//Is the year in 12-hour format?
	{
		*isHourPM = readData & MCP794_AM_PM;	//Mask out the AM/PM
		*getHour = bcdToDec(readData & MCP794_MASK_12Hour);	//Mask out the 12-hour time, convert from BCD and store in pointer
	}
	else
	{
		*isHourPM = 0;
		*getHour = bcdToDec(readData & MCP794_MASK_24Hour);	//Mask out the 24-hour time, convert from BCD and store in pointer
	}


	//---Minutes
	#if DEBUGLEVEL > 2
	UART_writeString("---Minutes---\r\n");
	#endif
	result = I2C_sendStart();					//Repeated Start - now we're going to do a Write operation
	result = I2C_send(deviceAddress|TW_WRITE);  //Send the device Address with WRITE
	result = I2C_send(MCP794_RTCMIN);			//Send Minute Register Address - to read from
	result = I2C_sendStart();					//Restart - now we're going to do a read operation
	result = I2C_send(deviceAddress|TW_READ);	//Send the device Address with READ
	readData = I2C_read(0);						//Read the value from the register
	*getMinutes = bcdToDec(readData);				//Convert from BCD and store value in pointer
	

	//---Seconds
	#if DEBUGLEVEL > 2
	UART_writeString("---Seconds---\r\n");
	#endif
	result = I2C_sendStart();					//Repeated Start - now we're going to do a Write operation
	result = I2C_send(deviceAddress|TW_WRITE);  //Send the device Address with WRITE
	result = I2C_send(MCP794_RTCSEC);			//Send Second Register Address - to read from
	result = I2C_sendStart();					//Repeated Start - now we're going to do a read operation
	result = I2C_send(deviceAddress|TW_READ);	//Send the device Address with READ
	readData = I2C_read(0);						//Read the value from the register
	*getSeconds = bcdToDec(readData & MCP794_MASK_Second);			//Mask out the Seconds only, convert from BCD and store value in pointer

	//Send STOP Condition
	I2C_sendStop();

	_delay_ms(1);

	return 1;	//Could implement error checking and make this a meaningful return value
}



/**
* @brief	Write a byte to the RTC
*
* @details	This function writes a byte to the RTC to a specified register
*
* @param[in]	deviceAddress	The I2C address of the RTC device
* @param[in]	registerAddress	The register to write to
* @pram[in]	data			The byte to be written to the register
*
* @return	1 = Success
************************************************************************/

uint8_t RTC_write(uint8_t deviceAddress, uint8_t registerAddress, uint8_t data)
{
	uint8_t result;
	
	//Send START Condition
	result = I2C_sendStart();
	
	//Send the device Address with WRITE
	result = I2C_send(deviceAddress|TW_WRITE);

	//Send Register Address to write to
	result = I2C_send(registerAddress);	

	//Write byte to register
	result = I2C_send(data);	//Data

	//Send STOP Condition
	I2C_sendStop();

	return result;
}



/**
* @brief	Read data from the RTC
*
* @details	This function reads data from the RTC from a specified register
*
* @param[in]	deviceAddress	The I2C address of the RTC device
* @param[in]	registerAddress	The register to read from
*
* @return	The byte read from RTC
************************************************************************/

uint8_t RTC_read(uint8_t deviceAddress, uint8_t registerAddress)
{
	
	uint8_t readData = 0;
	uint8_t result;


	//Send START Condition
	result = I2C_sendStart();

	//Send the device Address with WRITE - we need to write in order to specify the register to read from
	result = I2C_send(deviceAddress|TW_WRITE);

	//Send Register Address to read from
	result = I2C_send(registerAddress);	


	//Send RESTART Condition - now we read from the memory address
	result = I2C_sendStart();


	//Send the device Address with READ
	result = I2C_send(deviceAddress|TW_READ);

	//Read the data returned by the RTC
	readData = I2C_read(0);	//Read Data with NO ACK  (No ACK means we're done reading)

	//Send STOP Condition
	I2C_sendStop();
	
	//Return the data read
	return readData;
	
}


/**
* @brief	Convert Decimal to Binary-Coded Decimal
*
* @details	This function converts a decimal number to BCD
*
* @param[in]	val	The decimal number (max 99)
*
* @return	byte containing the BCD value
************************************************************************/
uint8_t decToBcd(uint8_t val)
{
	return ( (val / 10 * 16) + (val % 10) );
}


/**
* @brief	Convert Binary-Coded Decimal to Decimal
*
* @details	This function converts a BCD byte to a decimal number
*
* @param[in]	val	The decimal number (max 99)
*
* @return	byte containing the BCD value
************************************************************************/
uint8_t bcdToDec(uint8_t val)
{
	return ( (val / 16 * 10) + (val % 16) );
}
